
(*
 * Monad for type constructor M:
 *
 *
 * there are 2 combinators bind (>>=) : M 'a * ('a -> M 'b) -> M 'b
 * and return (aka unit) : 'a -> M 'a
 * satisfying the following laws:
 *
 * return is a kind of left identity for bind:    
 *    (return x) >>= f =~ f x
 * return is the right identity for bind:   
 *    mx >>= return =~ mx 
 * bind has a kind of associativity: 
 *    (mx >>= f) >>= g =~ mx >>= (fn y => (f y) >>= g))    
 *
 * 
 * these laws may be clearer if we define
 * Kleisli composition (>=>) : ('a -> M 'b) * ('b -> M 'c) -> (a -> M 'c)
 * using bind using the identity (f >=> g) = (fn x => f x >>= g)
 *
 * return is the left identity for Kleisli composition:
 *    return >=> f =~ f
 * return is the right identity for Kleisli composition:
 *    f >=> return =~ f
 * Kleisli composition is associative:
 *    f >=> (g >=> h) = (f >=> g) >=> h
 * 
 *) 


(*
 * return : 'a -> 'a option
 * REQUIRES: true
 * ENSURES: return satisfies the monad laws for 
 *  the type constructor M 'a = 'a option
 *)
val return :'a ->'a option = 
  SOME

(* 
 * bind : 'a option * ('a -> 'b option) -> 'b option
 * REQUIRES: true
 * ENSURES: bind satisfies the monad laws for
 *  the type constructor M 'a = 'a option
 *)
fun bind (mx:'a option, f:'a -> 'b option) : 'b option =
  Option.mapPartial f mx

(* infix alias of bind *)
infix >>= 
val op>>= = bind


(* 
 * kleisli : ('a -> 'b option) * ('b -> 'c option) -> 'a -> 'c option
 * REQUIRES: true
 * ENSURES: kleisli is Kleisli composition for bind 
 *)
fun kleisli (f:'a -> 'b option, g:'b -> 'c option) (x:'a) = 
  (f x) >>= g

(* infix alias of kleisli *)
infix >=> 
val op>=> = kleisli




fun inc n = n + 1
fun dec n = n - 1
fun dbl n = n * 2
fun half n = n div 2

fun plusN n = fn m => m + n 
fun timesN n = fn m => m * n

fun opN f n = fn m => f (m, n)
val minusN = opN op- 
val divN = opN (op div)


fun divNOpt n = 
  if n = 0 
  then fn _ => NONE 
  else fn m => SOME (m div n)


fun lift (f:'a -> 'b) : 'a -> 'b option = 
  return o f


infix |> 
fun x |> f = f x


(* the non-monad approach *)
val fullLift : ('a -> 'b) -> 'a option -> 'b option = 
  Option.map 